#
/*
 *		M U L T I - D E V I C E   Q U E U E I N G   S Y S T E M
 *
 *				   D Q L I S T S . C
 *
 *				Douglas P. Kingston III
 *
 *				Part of the MDQS Daemon
 */

#include <stdio.h>
#include "queue.h"

extern	struct queue *quehead;
extern	struct request *delhead;
extern	int delayupdate;
extern	char *Admprefix;

/*
 *		Q L I S T S ( )
 *
 *	This function produces the purely informational files
 *	which contain the current ordering of the requests
 *	as they are in the Daemon.  If these files cannot be made
 *	for some reason, it is not considered a fatal error.
 *	These are for the benefit of the q(I) program.
 */
qlists()
{
	char	tmpqfname[PATHSIZ];
	char	qfname[PATHSIZ];
	register struct request *rp;
	register struct queue *qp;
	FILE	*fp;
	int	fd;
	struct	qdata qdata;

	for( qp = quehead; qp != NULL; qp = qp->q_nextq ) {
		if( ! ( qp->q_status & Q_UPDATE ))
			continue;
		qp->q_status &= ~Q_UPDATE;
		sprintf( qfname, "%s/Q_%s", Admprefix, qp->q_name );
		sprintf( tmpqfname, "%s/TQ_%s", Admprefix, qp->q_name );

		unlink( tmpqfname );
		if(( fd = creat( tmpqfname, 0444 )) < 0 ) {
debug( "Cannot create %s (%s)", tmpqfname, sys_errlist[errno] );
			continue;
		}
		if(( fp = fdopen( fd, "w" )) == NULL ) {
			close( fd );
			unlink( tmpqfname );
			continue;
		}

		for( rp = qp->q_readyreq; rp != NULL; rp = rp->r_nextreq ) {
			qdata.qd_uid = rp->r_uid;
			qdata.qd_seqno = atoi( &rp->r_name[QN_SEQ] );
			fwrite( &qdata, sizeof( qdata ), 1, fp );
		}

		fflush( fp );
		if( ferror( fp ) == NULL ) {
			unlink( qfname );
			link( tmpqfname, qfname );
		} else
			debug( "Can't link to %s (%s)", qfname,
				sys_errlist[errno] );
		fclose( fp );
		unlink( tmpqfname );
	}

	if( delayupdate ) {
		delayupdate = 0;
		sprintf( qfname, "%s/DELAYED", Admprefix );
		sprintf( tmpqfname, "%s/TMPDELAYED", Admprefix );

		unlink( tmpqfname );
		if(( fd = creat( tmpqfname, 0444 )) < 0 ) {
debug( "Cannot create %s (%s)", tmpqfname, sys_errlist[errno] );
			return;
		}
		if(( fp = fdopen( fd, "w" )) == NULL ) {
			close( fd );
			unlink( tmpqfname );
			return;
		}

		for( rp = delhead; rp != NULL; rp = rp->r_nextreq ) {
			qdata.qd_uid = rp->r_uid;
			qdata.qd_seqno = atoi( &rp->r_name[QN_SEQ] );
			fwrite( &qdata, sizeof( qdata ), 1, fp );
		}

		fflush( fp );
		if( ferror( fp ) == NULL ) {
			unlink( qfname );
			link( tmpqfname, qfname );
		} else
			debug( "Can't link to %s (%s)", qfname,
				sys_errlist[errno] );
		fclose( fp );
		unlink( tmpqfname );
	}
}
