#
/*
 *	    M U L T I - D E V I C E   Q U E U E I N G   S Y S T E M
 *
 *			   Douglas P. Kingston III
 *
 *		     Lineprinter Server Process  (LPSERVER)
 *
 *	Note that a pipe to the Deamon will be on file descriptor 3.
 *	________________________________________________________________
 *
 *				Revision History
 *				================
 *	17 Sep 82  DPK	Modified copyn() to handle binary characters
 *			in a more intelligent fashion.  (\nnn notation)
 *
 *	16 Nov 82  DPK	Added required fclose in the routine print().
 *
 *	23 Dec 82  DPK	Altered handling of copyn to accomadate new copyn().
 */

#include <stdio.h>
#include <pwd.h>
#include <signal.h>

#if PRINTRONIX | VERSATEC | DCSERVER
#include <sgtty.h>
#endif
#if VERSATEC 
#include "vp.h"
#endif

#include "queue.h"
#include "qreplys.h"

time_t	timenow;
struct	request r;
int	lpoptions = 0;
long	linecount = 0L;	/* number of lines printed by this server */
char	source[LINESIZ]; /* source of request, account name or net address */
char	obuf[BUFSIZ];	/* remember the output goes to stdout */
char	*title = 0;
int	numcols = 132;
int	numhdrs = 1;
int	numtrailers = 0;
char	*arg0;		/* Pointer to the original argv[0] */
#if PRINTRONIX
struct sgttyb	ttyb =
{
	B1200,B1200,
	0, 0,
	CRMOD|XTABS|ANYP,
};
#endif
#if DCSERVER
#ifndef BRLUNIX
struct sgttyb ttyb =	/* tty line setup commands */
{
	B9600,B9600,
	0,0,
	CRMOD|XTABS|ANYP,
};
#else
struct sg_brl ttyb =	/* tty line setup commands */
{
	B9600,B9600,
	DC3DC1|STALL,
	CRMOD|XTABS|ANYP,
};
#endif BRLUNIX
#endif

struct	request *gethdr();
struct	passwd *getpwuid();
char	*index();
char	*ctime();
char	*strdup();
int	doflush();
int	dorestart();
long	copyn();	/* copyn() returns the number of lines output */

extern	char *Sitefile;
extern	char *Qdatadir;
#if VERSATEC
extern int	vprtmode[];
#endif

#define	FF	014	/* A formfeed character */

/*
 *		M A I N  ( )
 *  As a MDQS queue server, this process will be started up
 *  with the queue control file as the standard input, and
 *  the device of interest on the standard output.
 */
main (argc, argv)
int	argc;
char	**argv;
{
	char	datafile[LINESIZ];
	char	linebuf[LINESIZ];
	char	msg[LINESIZ];
	char	*cp;
	
	signal( SIGHUP, SIG_IGN );		/* Paranoid. */
	signal( SIGEMT, dorestart );
	signal( SIGTERM, doflush );

	arginit(argc, argv);
	time( &timenow );
	setbuf( stdout, obuf );
#if VERSATEC
	ioctl( fileno( stdout ), SETSTATE, vprtmode );
#endif
#if PRINTRONIX | DCSERVER
#ifdef BSD42
	bsd42tty(fileno(stdout));
#endif BSD42
	stty (fileno(stdout), &ttyb );
#endif

	if (gethdr (stdin, &r) == (struct request *) NULL) {
		err( "%s: Error reading request", arg0);
		retmsg (RP_FATAL, "Error reading request header");
		exit (10);
	}

	while (fgets (linebuf, sizeof (linebuf), stdin) != NULL) {
		if ((cp = index (linebuf, '\n')) != 0)
			*cp = 0;
		switch (linebuf[0]) {
		case 'B':
			statmsg ("printing header");
			sprintf (datafile, "%s/%s", Qdatadir, &linebuf[1]);
			prheader (datafile, &r);
			unlink (datafile);
			break;

		case 'D':
			if( chdir( &linebuf[1] ) < 0 ) {
				printf("\n\nError trying chdir (%s,%s)\n%c",
				&linebuf[1], sys_errlist[errno], FF );
				chdir( "/tmp" );	/* A safe default */
			}
			break;

		case 'U':
			strcpy( source, &linebuf[1] );
			break;

		case 'O':
			setoptions (&linebuf[1]);
			break;

		case 'H':
			statmsg("printing header");
			prheader (&linebuf[1], &r);
			break;

		case 'I':
			sprintf (datafile, "%s/%s", Qdatadir, &linebuf[1]);
			sprintf (msg, "printing %s", title);
			statmsg (msg);
			print (datafile);
			unlink (datafile);
			break;

		case 'F':
			sprintf (msg, "printing %s", &linebuf[1]);
			statmsg (msg);
			if (print (&linebuf[1]) < 0)
				break;
			if (lpoptions & NOKEEP)
				unlink (&linebuf[1]);
			break;

		case 'T':
			if (title) free(title);
			title = strdup(&linebuf[1]);
			break;

		default:
			/* Ignore unknown lines */
			break;
		}
	}
	if(numtrailers)
		prtrailer (&r);
	retmsg (RP_OK, 0);
	exit (0);
}

/*
 *		S E T O P T I O N S  ( )
 *  Setoptions is passed the ascii string numerical representation
 *  of the lpoptions flags word.  The bit definitions are in queue.h.
 *  The string is a decimal representation.
 */
setoptions (optstr)
char *optstr;
{
	lpoptions |= atoi (optstr);
}

/*
 *		P R I N T  ( )
 *  Dump a file onto the lineprinter.  We check for binary and insert
 *  spaces at the appropriate places.
 */
print( file )
char *file;
{
	FILE	*df;
	int	cnt;

	if ((df = fopen (file, "r")) == NULL) {
		printf ("\n\n***** Could not open input file (%s, %s) *****\n",
			file, sys_errlist[errno] );
		putchar( FF );
		return( -1 );
	}
	for( cnt = r.r_copies; cnt > 0; cnt--) {
		linecount += copyn (df, (r.r_size != 0L
					  ? r.r_size - linecount : 0L));
		if (r.r_size > 0 && linecount > r.r_size) {
			printf ("\n----- Printing limit of %ld lines exceeded, printing aborted. -----\n", r.r_size);
			break;
		}
		if( cnt > 1 ) {
			if (!(lpoptions & NOFF))
				putchar( FF );
			rewind( df );
		}
	}
	fclose( df );
	putchar( FF );
	return( 0 );
}

/*
 *	A R G I N I T
 *
 *	Parse args of lineprinter like servers
 */
arginit(argc, argv)
int	argc;
char	**argv;
{
	arg0 = argv[0];

	while(*++argv) {
		/*
		 *  Number of columns argument
		 */
		if (strncmp(*argv, "columns=", 8) == SAME) {
			numcols = atoi(*argv+8);
			if( numcols < 1 ) {
				numcols = 132;
				err("bad argument '%s'", *argv);
			}
			continue;
		}
		/*
		 *  Number of headers argument
		 */
		if (strncmp(*argv, "nheaders=", 9) == SAME) {
			numhdrs = atoi(*argv+9);
			if( numhdrs < 0 ) {
				numhdrs = 1;
				err("bad argument '%s'", *argv);
			}
			continue;
		}
		/*
		 *  Number of trailers argument
		 */
		if (strncmp(*argv, "ntrailers=", 10) == SAME) {
			numtrailers = atoi(*argv+10);
			if( numtrailers < 0 ) {
				numtrailers = 0;
				err("bad argument '%s'", *argv);
			}
			continue;
		}
		/*
		 *  Set NOFF
		 */
		if (strcmp(*argv, "noff") == SAME) {
			lpoptions |= NOFF;
			continue;
		}
#ifdef DCSERVER
		/*
		 *  Set line speed (override default of 9600)
		 */
		if (strncmp(*argv, "speed=", 6) == SAME) {
			int	baud;

			if ((baud = atoi (*argv+6)) == 0) {
				err ("Bad argument '%s'", *argv);
				continue;
			}
			switch (baud) {
			case 134:   ttyb.sg_ispeed = ttyb.sg_ospeed = B134; break;
			case 150:   ttyb.sg_ispeed = ttyb.sg_ospeed = B150; break;
			case 300:   ttyb.sg_ispeed = ttyb.sg_ospeed = B300; break;
			case 600:   ttyb.sg_ispeed = ttyb.sg_ospeed = B600; break;
			case 1200:  ttyb.sg_ispeed = ttyb.sg_ospeed = B1200; break;
			case 1800:  ttyb.sg_ispeed = ttyb.sg_ospeed = B1800; break;
			case 2400:  ttyb.sg_ispeed = ttyb.sg_ospeed = B2400; break;
			case 4800:  ttyb.sg_ispeed = ttyb.sg_ospeed = B4800; break;
			case 9600:  ttyb.sg_ispeed = ttyb.sg_ospeed = B9600; break;
			default:
				err("Unknown speed %d", baud);
				break;
			}
			continue;
		}
#endif DCSERVER
	}
}

/*
 *		E R R ( )
 *	Called by banner() and its subfunctions, amongst others
 */
/* VARARGS */
err( fmt, a, b, c, d )
char *fmt;
{
	fprintf( stderr, "%s: ", arg0);
	fprintf( stderr, fmt, a, b, c, d );
	putc( '\n', stderr );
}

#ifdef DCSERVER
#ifdef BSD42
bsd42tty(fd)
int	fd;
{
	int flag = LDECCTQ;

	ioctl (fd, TIOCSETD, NTTYDISC);
	ioctl (fd, TIOCLBIS, &flag);
}
#endif BSD42
#endif DCSERVER
