#
/*
 *	    M U L T I - D E V I C E   Q U E U E I N G   S Y S T E M
 *
 *			   Douglas P. Kingston III
 *
 *		     Request modifictation program (QMOD)
 *
 *				Revision History
 *				================
 *	6 Dec 82   DPK	Added xmalloc routine called by locking code.
 *
 *	30 Mar 83  DPK	Added newline to copies error message.  Thanks
 *			to Dave Wilson at Boeing.
 */

#include <stdio.h>
#include <errno.h>
#include "queue.h"
#include <pwd.h>
#include "dirent.h"

#ifdef DIRBLKSIZ		/* 4BSD flavor assumed */
#define	dirent	direct
#endif

#define	C_TSTART	000001
#define	C_COPIES	000002
#define	C_FORMS		000004
#define	C_OPTIONS	000010
#define	C_PRIORITY	000020
#define	C_QUEUE		000040
#define	C_STATUS	000100
#define C_REQUEST	000200

extern	int	Cp_min, Cp_max;
extern	int	Pri_min, Pri_max;
extern	char	*Qcntrldir;
extern	char	*Modprefix;
extern	char 	*Qdatadir;
extern	char	*Tmpqdir;

char	missingfmt[]	= "Missing %s argument";
char	dupfmt[]	= "Duplicate %s argument given";
int	error		= 0;
int	realuid;

int	changes		= 0;
time_t	starttime	= 0L;
int	copies		= 0;
int	priority	= 0;
char	*forms		= 0;
char	*queue		= 0;
int	qoptions	= 0;
int	notqoptions	= 0;
int	qstatus		= 0;
int	notqstatus	= 0;

extern time_t	getdate();
extern char	*index();
extern char	*ckforms();
extern struct	request *gethdr();
extern char	**findque();

/*
 *		M A I N ( )
 */
main (argc, argv)
int	argc;
char	**argv;
{
	register char	*arg;
	int	negate;
	register int	i;

	fdverify();			/* Paranoid */

	if( argc < 2 )
		usage();

	umask( 077 );

	if( chdir( Qcntrldir ) < 0 )
		bomb( "Cannot chdir to Qcntrldir\n" );

	/*
	 *   All queue modifications are done non-privileged.
	 */
	realuid = getuid();
	if (!mdqsmanager())
		setuid( realuid );

	for (i = 1; i < argc; i++) {
		if ( *argv[i] != '-' )
			break;
		arg = argv[i];
		arg++;
		negate = 0;
		if( *arg == '-' ) {
			negate++;
			arg++;
		}
		while (*arg) {
			switch (*arg) {
			case 'a':	/* Run after time */
				if (argv[i+1] == 0) {
					err (missingfmt, "time");
					error++;
					break;
				} else
					i++;
				if (starttime != 0L) {
					err (dupfmt, "time");
					error++;
				} else if ((starttime = getdate(argv[i], NULL)) == -1L) {
					err ("Bad time specification");
					error++;
				}
				changes |= C_TSTART;
				break;

			case 'c':	/* Number of copies desired */
				if (argv[i+1] == 0) {
					err (missingfmt, "copies");
					error++;
					break;
				} else
					i++;
				if (copies != 0) {
					err (dupfmt, "copies");
					error++;
				} else if (checknum(argv[i],
				    &copies, Cp_min, Cp_max) == -1) {
					err(
				     "Copies must be in the range %d to %d",
					  Cp_min, Cp_max);
					error++;
				}
				changes |= C_COPIES;
				break;

			case 'd':	/* Kill all reqs. for given queue */

				if (argv[i+1] == 0)   /* No queue name */
				{
					err(missingfmt, "queue");
					error++;
					break;
				}
				else
					i++;
				if (queue != 0)		/* Already have queue name */
				{
					err(dupfmt, "queue");
					error++;
				}
				else if (findque(argv[i]) == NULL)
				{
					err("Unknown queue %s", argv[i]);
					error++;
				}
				queue = argv[i];
				changes |= C_REQUEST;
				qstatus |= SKILL;
				break;

			case 'f':	/* What forms to print on */
				if (argv[i+1] == 0) {
					err (missingfmt, "forms");
					error++;
					break;
				} else
					i++;
				if (forms != 0) {
					err (dupfmt, "forms");
					error++;
				} else if ((forms = ckforms (argv[i])) == NULL) {
					err("Unknown form %s", argv[i]);
					error++;
				}
				changes |= C_FORMS;
				break;

			case 'h':	/* Temporarily hold the request */
				if( negate )
					notqstatus |= SHOLD;
				else
					qstatus |= SHOLD;
				changes |= C_STATUS;
				break;

			case 'k':	/* Kill the request! */
				if( negate )
					notqstatus |= SKILL; /* Good luck! */
				else
					qstatus |= SKILL;
				changes |= C_STATUS;
				break;

			case 'm':	/* Send mail upon completion */
				if( negate )
					notqoptions |= O_MAIL;
				else
					qoptions |= O_MAIL;
				changes |= C_OPTIONS;
				break;

			case 'p':
				if (argv[i+1] == 0) {
					err (missingfmt, "priority");
					error++;
					break;
				} else
					i++;
				if (priority != 0) {
					err (dupfmt, "priority");
					error++;
				} else if (checknum(argv[i],
				    &priority, Pri_min, Pri_max) == -1) {
					err(
				  "Priorities must be in the range %d to %d",
					  Pri_min, Pri_max);
					error++;
				}
				changes |= C_PRIORITY;
				break;

			case 'q':
				if (argv[i+1] == 0) {
					err (missingfmt, "queue");
					error++;
					break;
				} else
					i++;
				if (queue != 0) {
					err (dupfmt, "queue");
					error++;
				} else if( findque( argv[i] ) == NULL ) {
					err ("Unknown queue '%s'", argv[i]);
					error++;
				}
				queue = argv[i];
				changes |= C_QUEUE;
				break;

			case '?':
				error++;
				break;

			default:
				err ("Unknown flag %c", *arg);
				error++;
			}
			arg++;
		}
	}
	
	if( error )
		usage();

	if (changes & C_REQUEST)
	{
		procs2(queue);
		proddaemon();
		exit(0);
	}

	for( ; i < argc; i++ ) {
		process( argv[i] );
	}
	proddaemon();
	exit (0);
}

usage()
{
	err ("Usage:   qmod [options] requests");
	err ("Options: [-a time] [-c #] [-d que] [-f form]");
	err ("         [-[-]hkm] [-p pri] [-q que]");
	exit (1);
}

/*
 *	P R O C E S S ( )
 *	Fix one request.
 */
process( arg )
register char	*arg;
{
	struct passwd *pw;
	struct passwd *getpwnam();
	struct request r;
	register char	*cp;
	char	rname[QDIRSIZ+1];	/* Name of the controlf file */
	char	mname[PATHSIZ];		/* buffer for building the Mod link */
	int	uid;
	int	rfd;			/* FD of the control file */
	FILE	*rfp;			/* file pointer for above */
	int	reqno;

	if( cp = index( arg, '.' )) {
		*cp++ = 0;
		uid = atoi( arg );
		if( uid == 0 && arg[0] != '0' ) {
			if(( pw = getpwnam( arg )) != NULL )
				uid = pw->pw_uid;
			else {
				err( "Unknown userid '%s'.", arg );
				return;
			}
		}
		reqno = atoi( cp );
	} else {
		uid = realuid;
		reqno = atoi( arg );
	}
	sprintf(rname, "Q%05.5d.%d", uid, reqno );
	if(( rfd = open( rname, READ )) < 0 ) {
		if( errno == ETXTBSY )
			err( "Request %s locked, try again", rname );
		else
			err( "Cannot find request %s (%s)", rname,
			      sys_errlist[errno] );
		return;
	}

	rfp = fdopen(rfd, "r");
	if( gethdr (rfp, &r) == NULL ) {
		err( "Error reading request %s (%s)", rname,
		      sys_errlist[errno] );
		fclose( rfp );
		return;
	}
	fclose( rfp );

	if( changes & C_TSTART )
		r.r_tstart = ( starttime < r.r_tsubmit ) ?
				r.r_tsubmit : starttime;

	if( changes & C_COPIES )
		r.r_copies = copies;

	if( changes & C_FORMS )
		strncpy (r.r_forms, forms, FORMSIZ);

	if( changes & C_OPTIONS ) {
		r.r_options |= qoptions;
		r.r_options &= ~notqoptions;
	}

	if( changes & C_QUEUE )
		strncpy (r.r_que, queue, QUESIZ);

	if( changes & C_PRIORITY )
		r.r_pri = priority;

	if( changes & C_STATUS ) {
		r.r_status |= qstatus;
		r.r_status &= ~notqstatus;
	}
	/*
	 *  Now indicate that it has changed to the daemon, write
	 *  out the request structure into new file in Mod directory
	 */
	sprintf( mname, "%s/%s", Modprefix, rname );
	if(( rfd = creat( mname, 0400 )) < 0 )
	 {
		if( errno == EACCES )
			err( "Previous change still pending, try again" );
		else
			err( "Could not create modification file for %s (%s)",
				mname, sys_errlist[errno] );
		return;
	}
	rfp = fdopen (rfd, "w");
	puthdr (&r, rfp);
	fflush (rfp);
	if(ferror (rfp))
		err( "Error writing changed request (%s)", sys_errlist[errno]);
	fclose( rfp );
	return;
}




/*
 *		P R O C E S S 2 ( ) 
 *  Opens a q control file and then calls modifyque()
 *
 */
procs2(queue)
char *queue;
{
	DIR	*df;
	struct	dirent *dp;

	if ((df = opendir(".")) == NULL)
		bomb("Can't open %s/. (%s)\n", Qcntrldir, sys_errlist[errno]);
	while ((dp = readdir(df)) != NULL)
	{
		if (dp->d_name[0] == 'Q')    /* A control file */
			modifyque(dp->d_name, queue);
	}
	return;
}




/*
 *		M O D I F Y Q U E ( ) 
 *  Checks all requests if for given queue,  then
 *  mark requests as deleted.
 *
 */

modifyque(cfname, queue)
char *cfname;
char *queue;
{
	FILE *cf;
	int cfd;
	struct request r;
	char mname[PATHSIZ];	/* buffer for build Mod link */

	if ((cf = fopen(cfname, "r")) == NULL)
		return;

	
	if (gethdr(cf, &r) == (struct request *) NULL)
	{
		err("Error reading request structure !");
		fclose(cf);
		return;
	}

	if (queue && strncmp(r.r_que, queue, QUESIZ) != SAME)
	{		/* DON'T KILL THIS ONE */
		fclose(cf);
		return;
	}

	fclose(cf);
	r.r_status |= qstatus;
		
	/*
	 *  Now indicate that it has changed to the daemon, write
	 *  out the request structure into new file in Mod directory
	 */
	sprintf( mname, "%s/%s", Modprefix, cfname );
	if(( cfd = creat( mname, 0400 )) < 0 )
	 {
		if( errno == EACCES )
			err( "Previous change still pending, try again" );
		else
			err( "Could not create modification file for %s (%s)",
				mname, sys_errlist[errno] );
		return;
	}
	cf = fdopen (cfd, "w");
	puthdr (&r, cf);
	fflush (cf);
	if(ferror (cf))
		err( "Error writing changed request (%s)", sys_errlist[errno]);
	fclose( cf );
	return;
}


/*
 *		C H E C K N U M ( )
 *  Make sure the number given is within certain bounds.  The number
 *  is given in ascii.  -1 is returned if the number is out of bounds
 *  and 0 otherwise.  The value of the signed string number is placed
 *  in the integer pointed at by valuep.
 */
checknum (number, valuep, min, max)
char	*number;
int	*valuep;
int	min;
int	max;
{
	int	val;
	
	if ((val = atoi (number)) == 0 && *number != 0)
		return (-1);
	if (val < min || val > max)
		return (-1);
	*valuep = val;
	return (0);
}
